/* ----------------------------------------------------------
(c) 2007 Sylvain Lefebvre - all rights reserved
--
The source code is provided 'as it is', without any warranties. 
Use at your own risk. The use of any part of the source code in a
commercial or non commercial product without explicit authorisation
from the author is forbidden. Use for research and educational
purposes is allowed and encouraged, provided that a short notice
acknowledges the author's work.
---------------------------------------------------------- */
// ------------------------------------------------------
//
// Image class
// 
// Coordinate system origin is at top left corner
// pixel(x,y) retrieves pixel at row y, column x  
// ( or Array2D::get(x,y) )
//
// ------------------------------------------------------
// Sylvain Lefebvre - 2006-03-07
// ------------------------------------------------------

#pragma once

#include "Tuple.h"

#include <map>
#include <string>

/// Base Image class
class Image
  //  : public LibSL::Memory::TraceLeaks::LeakProbe<Image>
{
public:

  virtual ~Image() {}

  virtual uint   w()         const =0;
  virtual uint   h()         const =0;
  virtual uint   numComp()   const =0;
  virtual uchar*       raw()       =0;
  virtual const uchar* raw() const =0;
};


/// Image format plugin (abstract)
class ImageFormat_plugin
{
public:
  virtual void        save(const char *,const Image *) const =0;
  virtual Image      *load(const char *)               const =0;
  virtual const char *signature()                      const =0;
  virtual ~ImageFormat_plugin() {}
};


/// Image format manager (singleton)
class ImageFormatManager
{
private:

  std::map<std::string,ImageFormat_plugin*> m_Plugins;

  static ImageFormatManager *s_Manager;
  ImageFormatManager();

public:

  ~ImageFormatManager();

  void registerPlugin(ImageFormat_plugin *plugin);
  ImageFormat_plugin *getPlugin(const char *signature) const;

  static ImageFormatManager *getUniqueInstance();

};


/// Generic Image class
template <typename T_Type,unsigned int T_NumComp>
class Image_generic : public Image
{
public:

  typedef T_Type                            t_Component;
  typedef Tuple<T_Type,T_NumComp>           t_Pixel;

  enum {e_NumComp = T_NumComp};

protected:

  t_Pixel *m_PixelArray;
  uint     m_W;
  uint     m_H;

public:

  Image_generic()
  {
    m_PixelArray = NULL;
    m_W          = 0;
    m_H          = 0;
  }

  Image_generic(uint w, uint h)
  {
    m_PixelArray = new t_Pixel[w*h];
    m_W          = w;
    m_H          = h;
    memset(m_PixelArray,0,sizeof(t_Pixel)*w*h);
  }

  Image_generic(const Image_generic& img)
  {
    if (m_PixelArray != NULL) {
      delete [](m_PixelArray);
    }
    m_PixelArray = new t_Pixel[img.w()*img.h()];
    m_W          = img.w();
    m_H          = img.h();
    memcpy(m_PixelArray,img.m_PixelArray,sizeof(t_Pixel)*img.w()*img.h());
  }

  ~Image_generic()
  {
    if (m_PixelArray != NULL) {
      delete [](m_PixelArray);
    }
  }

  /// Accessors to single pixel
  const t_Pixel& pixel(uint i, uint j) const { return m_PixelArray[(i+j*w())]; }
  t_Pixel&       pixel(uint i, uint j)       { return m_PixelArray[(i+j*w())]; }

  /// Image sizes
  uint w() const       {return (m_W);}
  uint h() const       {return (m_H);}
  uint numComp() const {return (T_NumComp);}

  /// low level memory access
  uchar*        raw()       {return reinterpret_cast<uchar*      >(m_PixelArray);}
  const uchar*  raw() const {return reinterpret_cast<const uchar*>(m_PixelArray);}

  /// Copy an image at the given location
  void copy(int x,int y,const Image_generic *img)
  {
    ForImage(img,i,j) {
      int xd=i+x;
      int yd=j+y;
      if ((xd >= 0 && xd < int(w())) &&  (yd >= 0 && yd < int(h()))) {
        pixel(xd,yd)=img->pixel(i,j);
      }
    }
  }

  /// Extract a sub-image
  Image_generic *extract(int xe,int ye,uint we,uint he)
  {
    sl_assert(we > 0 && he > 0);
    Image_generic *sub = new Image_generic(we,he);
    ForImage(sub,i,j) {
      int xs = xe+i;
      int ys = ye+j;
      if (xs < 0 || xs >= int(w())) continue;
      if (ys < 0 || ys >= int(h())) continue;
      sub->pixel(i,j) = pixel(xs,ys);
    }
    return (sub);
  }

  /// Cast into another image type
  template <class T_Image> typename T_Image *cast() const
  {
    typename T_Image *nimg = new T_Image(w(),h());
    ForImage(nimg,i,j) {
      ForIndex(c,min(int(e_NumComp),int(T_Image::e_NumComp))) {
        nimg->pixel(i,j)[c] = (typename T_Image::t_Component)pixel(i,j)[c];
      }
    }
    return (nimg);
  }

};

/// Load and save global methods

Image    *loadImage_raw(const char *);

/// Load an image and cast it

template<class T_Image> typename T_Image*
loadImage(const char *fname)
{
  Image   *img = loadImage_raw(fname);
  T_Image *ptr = dynamic_cast<T_Image *>(img);
  if (ptr == NULL) {
    delete (img); // delete image!
    fprintf(stderr,"loadImage<> - Image format mismatch! (%s)",fname);
    exit (-1);
  }
  return (ptr); 
}

typedef Image_generic<unsigned char,3>            ImageRGB;
typedef Image_generic<unsigned char,4>            ImageRGBA;

// ------------------------------------------------------

#define IMAGE_FORMAT_MANAGER (*ImageFormatManager::getUniqueInstance())

// ------------------------------------------------------
