//--------------------------------------------------------------------------------------
/*

Demo code for 'Filtered Tilemaps' - ShaderX6

Requirements:
-------------

- DX10 hardware
- DirectX SDK must be installed
- DXUT must be compiled ( $(DXSDK_DIR)\Samples\C++\DXUT\Core\ and $(DXSDK_DIR)\Samples\C++\DXUT\Optional\ )

--

(c) Sylvain Lefebvre - 2007-03-29

--

Code status: prototype

--

The source code is provided 'as it is', without any warranties. 
Use at your own risk. Permission to reuse and modify the code 
from this file is granted for any purpose.

*/
//--------------------------------------------------------------------------------------

#include "DXUT.h"
#include "DXUTcamera.h"
#include "DXUTgui.h"
#include "DXUTsettingsDlg.h"
#include "SDKmisc.h"
#include "SDKMesh.h"
#include "resource.h"

#include "Trackball.h"       // a simple trackball
#include "Image.h"           // an image class
#include "ImageFormat_PNG.h" // PNG loader
ImageFormat_PNG s_PNG;

using namespace std;

//--------------------------------------------------------------------------------------

#define NUM_COLORS    2                                             // number of edge colors
#define NUM_TILES     (NUM_COLORS*NUM_COLORS*NUM_COLORS*NUM_COLORS) // number of tiles
#define TILEMAP_SIZE  128                                           // tilemap size

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------

// UI
CDXUTDialogResourceManager  g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg             g_D3DSettingsDlg;        // device settings dialog
CDXUTDialog                 g_HUD;                   // manages the 3D   
CDXUTDialog                 g_SampleUI;              // dialog for sample specific controls
CDXUTTextHelper            *g_pTxtHelper   = NULL;
CDXUTListBox               *g_lbRenderMode = NULL;   // select render mode
CDXUTListBox               *g_lbTileSet    = NULL;   // select tile set
// Font
ID3DX10Font*                g_pFont10       = NULL;       
ID3DX10Sprite*              g_pSprite10     = NULL;
// Effect
ID3D10Effect*               g_pEffect10              = NULL;
ID3D10EffectTechnique*      g_RenderTMTechnique      = NULL;
ID3D10EffectTechnique*      g_RenderITMTechnique     = NULL;
ID3D10EffectTechnique*      g_RenderITMCTechnique    = NULL;
ID3D10EffectTechnique*      g_RenderPTMTechnique     = NULL;
ID3D10EffectTechnique*      g_SelectedTechnique      = NULL;
ID3D10EffectMatrixVariable *g_fxTrackball            = NULL;
ID3D10EffectMatrixVariable *g_fxProj                 = NULL;
// Screen width / height
int                         g_Width         = 1024;
int                         g_Height        = 768;
// Tile width (assumes square tiles)
float                       g_TileW         = 0;
// Quad VB/IB/VL
ID3D10Buffer*               g_QuadVB        = NULL;
ID3D10Buffer*               g_QuadIB        = NULL;
ID3D10InputLayout*          g_QuadVL        = NULL;
// Packed tiles texture
ID3D10Texture2D*            g_Tex2D         = NULL;
ID3D10ShaderResourceView   *g_Tex2DRV       = NULL;
// Array of tiles
ID3D10Texture2D*            g_Array2D       = NULL;
ID3D10ShaderResourceView   *g_Array2DRV     = NULL;
// Tilemap
ID3D10Texture2D*            g_TileMap       = NULL;
ID3D10ShaderResourceView   *g_TileMapRV     = NULL;
// Prefiltered tilemap
ID3D10Texture2D*            g_PFTileMap     = NULL;
ID3D10ShaderResourceView   *g_PFTileMapRV   = NULL;
// Blend state
ID3D10BlendState*           g_StateNoBlend  = NULL;

//--------------------------------------------------------------------------------------

// a quad vertex
struct quad_vertex {
  D3DXVECTOR3 pos;
  D3DXVECTOR2 uv;
};

//--------------------------------------------------------------------------------------

// Macro for error reporting
#define CHECKERR(x) { HRESULT hr = (x); if( FAILED(hr) ) { MessageBoxA(NULL, #x, "DX10 Error",MB_OK|MB_ICONERROR); DXUTShutdown(); exit (-1);} }

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------

#define IDC_TOGGLEFULLSCREEN    1
#define IDC_TOGGLEREF           2
#define IDC_CHANGEDEVICE        3

#define IDC_RELOAD_SHADER       4
#define IDC_RENDER_MODE         5
#define IDC_TILE_SET            6

//--------------------------------------------------------------------------------------
// Load the shader, setup the UI
//--------------------------------------------------------------------------------------

HRESULT loadShader(ID3D10Device* pd3dDevice)
{
  SAFE_RELEASE( g_pEffect10 );

  // Read the D3DX effect file
  WCHAR str[MAX_PATH];
  CHECKERR( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"dx10tiling.fx" ) );
  CHECKERR( D3DX10CreateEffectFromFile(
    str, NULL, NULL, "fx_4_0",
    0, 0,
    pd3dDevice, NULL, NULL, &g_pEffect10, NULL, NULL ) );

  // Get handle on variables
  g_fxTrackball           = g_pEffect10->GetVariableByName( "Trackball" )->AsMatrix();
  g_fxProj                = g_pEffect10->GetVariableByName( "Proj" )->AsMatrix();

  // Get rendering techniques
  g_RenderTMTechnique      = g_pEffect10->GetTechniqueByName( "RenderTileMap" );
  g_RenderITMTechnique     = g_pEffect10->GetTechniqueByName( "RenderInterpolatedTileMap" );
  g_RenderITMCTechnique    = g_pEffect10->GetTechniqueByName( "RenderInterpolatedTileMapConditional" );
  g_RenderPTMTechnique     = g_pEffect10->GetTechniqueByName( "RenderPackedTileMap" );
  
  g_SelectedTechnique      = g_RenderITMTechnique;

  // Setup UI
  g_lbRenderMode->RemoveAllItems();
  g_lbRenderMode->AddItem(L"No interpolation"      ,g_RenderTMTechnique);
  g_lbRenderMode->AddItem(L"Interpolation"         ,g_RenderITMTechnique);
  g_lbRenderMode->AddItem(L"Interpolation (cond.)" ,g_RenderITMCTechnique);
  g_lbRenderMode->AddItem(L"Packing"               ,g_RenderPTMTechnique);

  g_lbTileSet->RemoveAllItems();
  g_lbTileSet->AddItem(L"Pebbles" ,"4x4.pebbles.png");
  g_lbTileSet->AddItem(L"Leopard" ,"4x4.leopard.png");
  g_lbTileSet->AddItem(L"Candies" ,"4x4.candies.png");
  g_lbTileSet->AddItem(L"Peppers" ,"4x4.peppers.png");
  g_lbTileSet->AddItem(L"Pipes"   ,"4x4.pipes.png");
  g_lbTileSet->AddItem(L"Towel"   ,"4x4.towel.png");
  g_lbTileSet->AddItem(L"Borders" ,"4x4.borders.png");
  g_lbTileSet->AddItem(L"Flat"    ,"4x4.flat.png");

  return S_OK;
}


// ----------------------------------------------------------
// Set all shader parameters
// ----------------------------------------------------------

HRESULT setShaderParameters(ID3D10Device* pd3dDevice)
{

  // -> 2d texture array
  ID3D10EffectShaderResourceVariable *fxArray2D = g_pEffect10->GetVariableByName( "Array2d" )->AsShaderResource();
  CHECKERR ( fxArray2D->SetResource(g_Array2DRV) );
  // -> packed tiles texture
  ID3D10EffectShaderResourceVariable *fxTex2D   = g_pEffect10->GetVariableByName( "Packed" )->AsShaderResource();
  CHECKERR ( fxTex2D->SetResource(g_Tex2DRV) );
  // -> tile width
  ID3D10EffectVariable *fxTileSize = g_pEffect10->GetVariableByName( "TileWidth" );
  fxTileSize   ->AsScalar()->SetFloat(g_TileW);
  ID3D10EffectVariable *fxInvTileSize = g_pEffect10->GetVariableByName( "TileWidth_inv" );
  fxInvTileSize->AsScalar()->SetFloat(1.0f / float(g_TileW));
  // -> tilemap
  ID3D10EffectShaderResourceVariable *fxTileMap = g_pEffect10->GetVariableByName( "TileMap" )->AsShaderResource();
  CHECKERR ( fxTileMap->SetResource(g_TileMapRV) );
  // -> pre-filtered tilemap
  ID3D10EffectShaderResourceVariable *fxPFTileMap = g_pEffect10->GetVariableByName( "PreFiltered" )->AsShaderResource();
  CHECKERR ( fxPFTileMap->SetResource(g_PFTileMapRV) );

  return S_OK;
}

// ----------------------------------------------------------
// Compute a tile index from edge colors (continuous packing [Wei2004])
// ----------------------------------------------------------

int tileIndex(int e1,int e2)
{
  if (e1 == 0 && e2 == 0) {
    return 0;
  }
  if  (e1 > e2 && e2 > 0) {
    return (e1*e1+2*e2-1);
  }
  if (e2 > e1 && e1 >= 0) {
    return 2*e1 + e2*e2;
  }
  if (e1 == e2 && e2 > 0) {
    return (e1+1)*(e1+1)-2;
  }
  if (e1 > e2 && e2 == 0) {
    return (e1+1)*(e1+1)-1;
  }
  return (-1);
}

//--------------------------------------------------------------------------------------
// Load the tiles and create all texture ressources
//--------------------------------------------------------------------------------------

void loadTiles(ID3D10Device* pd3dDevice,const char *fname)
{
  /// Load tile set
  ImageRGB  *tiles_rgb = loadImage<ImageRGB>(fname);
  ImageRGBA *tiles     = tiles_rgb->cast<ImageRGBA>();
  delete (tiles_rgb);
  // -> compute tile size and number of levels
  uint tile_w     = tiles->w() / (NUM_COLORS*NUM_COLORS);
  uint tile_h     = tiles->h() / (NUM_COLORS*NUM_COLORS);
  g_TileW         = float(tile_w);
  uint num_levels = 0;
  while ((1 << (num_levels)) <= int(tile_w)) {
    num_levels ++;
  }
  // -> create MIP-map pyramids for the tiles (simple box filter)
  vector<ImageRGBA*> pyr;
  pyr.resize(num_levels);
  pyr[0] = tiles;
  ForRange(l,1,pyr.size()-1) {
    pyr[l] = new ImageRGBA(max(1,tiles->w()>>l),max(1,tiles->h()>>l));
    ForImage(pyr[l],x,y) {
      v4f c00 = v4f(pyr[l-1]->pixel(x*2  ,y*2  ));
      v4f c01 = v4f(pyr[l-1]->pixel(x*2  ,y*2+1));
      v4f c11 = v4f(pyr[l-1]->pixel(x*2+1,y*2+1));
      v4f c10 = v4f(pyr[l-1]->pixel(x*2+1,y*2  ));
      v4f avg = (c00+c01+c10+c11)*0.25f;
      pyr[l]->pixel(x,y) = v4b(avg);
    }
  }

  /// Create packed tile texture (continuous packing)
  // -> data
  D3D10_TEXTURE2D_DESC             tex2dDesc;
  D3D10_SHADER_RESOURCE_VIEW_DESC  tex2dRVDesc;
  vector<D3D10_SUBRESOURCE_DATA>     texMipData;
  texMipData.resize(num_levels);
  for (uint l=0;l<num_levels;l++) {
    texMipData[l].pSysMem             = pyr[l]->raw();
    texMipData[l].SysMemPitch         = pyr[l]->w()*4;
    texMipData[l].SysMemSlicePitch    = 0;
  }
  // -> dx10 description
  tex2dDesc.ArraySize			      = 1;
  tex2dDesc.Format			        = DXGI_FORMAT_R8G8B8A8_UNORM;
  tex2dDesc.Width				        = tiles->w();
  tex2dDesc.Height			        = tiles->h();
  tex2dDesc.MipLevels			      = num_levels;
  tex2dDesc.MiscFlags			      = 0;
  tex2dDesc.SampleDesc.Count    = 1;
  tex2dDesc.SampleDesc.Quality  = 0;
  tex2dDesc.CPUAccessFlags	    = 0;
  tex2dDesc.Usage				        = D3D10_USAGE_DEFAULT;
  tex2dDesc.BindFlags			      = D3D10_BIND_SHADER_RESOURCE;
  SAFE_RELEASE(g_Tex2D);
  CHECKERR( pd3dDevice->CreateTexture2D( &tex2dDesc, &(texMipData[0]), &g_Tex2D ) );
  // -> dx10 view
  tex2dRVDesc.Format                    = DXGI_FORMAT_R8G8B8A8_UNORM;
  tex2dRVDesc.ViewDimension             = D3D10_SRV_DIMENSION_TEXTURE2D;
  tex2dRVDesc.Texture2D.MipLevels       = num_levels;
  tex2dRVDesc.Texture2D.MostDetailedMip = 0;
  SAFE_RELEASE(g_Tex2DRV);
  CHECKERR( pd3dDevice->CreateShaderResourceView( g_Tex2D, &tex2dRVDesc, &g_Tex2DRV ) );

  /// Create array of tiles (each tile has its own MIP-map pyramid)
  // -> data
  Tuple<uint,3> sz;
  sz[2] = NUM_TILES;
  sz[1] = num_levels;
  sz[0] = tile_w*tile_h*4;
  vector<vector<vector<uchar> > > tex2dData;
  tex2dData.resize(sz[2]);
  ForIndex(i,sz[2]) {
    tex2dData[i].resize(sz[1]);
    ForIndex(j,sz[1]) {
      tex2dData[i][j].resize(sz[0]);
    }
  }
  // -> organize tiles in an array
  for (uint k=0;k<NUM_TILES;k++) {
    for (uint l=0;l<num_levels;l++) {
      uint s = 1 << l;
      uint w = tile_w / s;
      uint h = tile_h / s;
      for (uint j=0;j<h;j++) {
        for (uint i=0;i<w;i++) {
          int ti  = (k%(NUM_COLORS*NUM_COLORS))*w;
          int tj  = (k/(NUM_COLORS*NUM_COLORS))*h;
          v4b pix = pyr[l]->pixel(ti+i,tj+j);
          tex2dData[k][l][(i+j*w)*4 + 0] = pix[0];
          tex2dData[k][l][(i+j*w)*4 + 1] = pix[1];
          tex2dData[k][l][(i+j*w)*4 + 2] = pix[2];
          tex2dData[k][l][(i+j*w)*4 + 3] = (w == 1 || h == 1) ? 0 : 255;
        }
      }
    }
  }
  vector<D3D10_SUBRESOURCE_DATA> texArrayData; 
  texArrayData.resize(NUM_TILES*num_levels);
  for (uint k=0;k<NUM_TILES;k++) {
    for (uint l=0;l<num_levels;l++) {
      texArrayData[k*num_levels+l].pSysMem             = &(tex2dData[k][l][0]);
      texArrayData[k*num_levels+l].SysMemPitch         = (tile_w >> l)*4;
      texArrayData[k*num_levels+l].SysMemSlicePitch    = 0;
    }
  }
  // -> dx10 description
  tex2dDesc.ArraySize			      = NUM_TILES;
  tex2dDesc.Format			        = DXGI_FORMAT_R8G8B8A8_UNORM; // unsigned bytes remapped to [0..1]
  tex2dDesc.Width				        = tile_w;
  tex2dDesc.Height			        = tile_h;
  tex2dDesc.MipLevels			      = num_levels;
  tex2dDesc.MiscFlags			      = 0;
  tex2dDesc.SampleDesc.Count    = 1;
  tex2dDesc.SampleDesc.Quality  = 0;
  tex2dDesc.CPUAccessFlags	    = 0;
  tex2dDesc.Usage				        = D3D10_USAGE_DEFAULT;
  tex2dDesc.BindFlags			      = D3D10_BIND_SHADER_RESOURCE;
  SAFE_RELEASE(g_Array2D);
  CHECKERR( pd3dDevice->CreateTexture2D( &tex2dDesc, &(texArrayData[0]), &g_Array2D ) );
  // -> dx10 view
  tex2dRVDesc.Format                         = DXGI_FORMAT_R8G8B8A8_UNORM;
  tex2dRVDesc.ViewDimension                  = D3D10_SRV_DIMENSION_TEXTURE2DARRAY;
  tex2dRVDesc.Texture2DArray.MipLevels       = num_levels;
  tex2dRVDesc.Texture2DArray.MostDetailedMip = 0;
  tex2dRVDesc.Texture2DArray.ArraySize       = NUM_TILES;
  tex2dRVDesc.Texture2DArray.FirstArraySlice = 0;
  SAFE_RELEASE(g_Array2DRV);
  CHECKERR( pd3dDevice->CreateShaderResourceView( g_Array2D, &tex2dRVDesc, &g_Array2DRV ) );

  /// Create tilemap
  // -> fill-in the tile map randomly
  // -> first, randomly select edges
  unsigned char edgeIds[(TILEMAP_SIZE+1)*(TILEMAP_SIZE+1)*2];
  memset(edgeIds,0,(TILEMAP_SIZE+1)*(TILEMAP_SIZE+1));
  for (int k=0;k<TILEMAP_SIZE+1;k++) {
    for (int l=0;l<TILEMAP_SIZE+1;l++) {
      edgeIds[(k+l*(TILEMAP_SIZE+1))*2+0] = uchar((rand()/113) % NUM_COLORS);
      edgeIds[(k+l*(TILEMAP_SIZE+1))*2+1] = uchar((rand()/ 17) % NUM_COLORS);
    }
  }
  // -> second, determine tile ids from edges
  unsigned char tileIds[TILEMAP_SIZE*TILEMAP_SIZE*4];
  memset(tileIds,0,TILEMAP_SIZE*TILEMAP_SIZE*4);
  for (int k=0;k<TILEMAP_SIZE;k++) {
    for (int l=0;l<TILEMAP_SIZE;l++) {
      int h0 = edgeIds[(l+ k   *(TILEMAP_SIZE+1))*2+0];
      int h1 = edgeIds[(l+(k+1)*(TILEMAP_SIZE+1))*2+0];
      int v0 = edgeIds[( l   +k*(TILEMAP_SIZE+1))*2+1];
      int v1 = edgeIds[((l+1)+k*(TILEMAP_SIZE+1))*2+1];
      int ti = tileIndex(v0,v1) + (NUM_COLORS*NUM_COLORS)*(tileIndex(h0,h1));
      tileIds[(l+k*TILEMAP_SIZE)*4+0] = uchar(ti);
    }
  }
  // -> fill-in data for faster lookup (store index of neighbors)
  for (int k=0;k<TILEMAP_SIZE-1;k++) {
    for (int l=0;l<TILEMAP_SIZE-1;l++) {
      tileIds[(l+k*TILEMAP_SIZE)*4+1] = tileIds[((l+1)+(k  )*TILEMAP_SIZE)*4+0];
      tileIds[(l+k*TILEMAP_SIZE)*4+2] = tileIds[((l  )+(k+1)*TILEMAP_SIZE)*4+0];
      tileIds[(l+k*TILEMAP_SIZE)*4+3] = tileIds[((l+1)+(k+1)*TILEMAP_SIZE)*4+0];
    }
  }
  // -> data
  D3D10_SUBRESOURCE_DATA initData;
  initData.pSysMem              = tileIds;
  initData.SysMemPitch          = TILEMAP_SIZE*4;
  initData.SysMemSlicePitch     = 0;
  // -> dx10 description
  tex2dDesc.ArraySize			      = 1;
  tex2dDesc.Format			        = DXGI_FORMAT_R8G8B8A8_UINT;
  tex2dDesc.Width				        = TILEMAP_SIZE;
  tex2dDesc.Height			        = TILEMAP_SIZE;
  tex2dDesc.MipLevels			      = 1;
  tex2dDesc.MiscFlags			      = 0;
  tex2dDesc.SampleDesc.Count    = 1;
  tex2dDesc.SampleDesc.Quality  = 0;
  tex2dDesc.CPUAccessFlags	    = 0;
  tex2dDesc.Usage				        = D3D10_USAGE_DEFAULT;
  tex2dDesc.BindFlags			      = D3D10_BIND_SHADER_RESOURCE;
  SAFE_RELEASE(g_TileMap);
  CHECKERR( pd3dDevice->CreateTexture2D( &tex2dDesc, &initData, &g_TileMap ) );
  // -> dx10 view
  tex2dRVDesc.Format                    = DXGI_FORMAT_R8G8B8A8_UINT;
  tex2dRVDesc.ViewDimension             = D3D10_SRV_DIMENSION_TEXTURE2D;
  tex2dRVDesc.Texture2D.MipLevels       = 1;
  tex2dRVDesc.Texture2D.MostDetailedMip = 0;
  SAFE_RELEASE(g_TileMapRV);
  CHECKERR( pd3dDevice->CreateShaderResourceView( g_TileMap, &tex2dRVDesc, &g_TileMapRV ) );
  
  /// Create pre-filtered tilemap
  ImageRGBA *prefiltered = new ImageRGBA(TILEMAP_SIZE*2,TILEMAP_SIZE*2);
  for (int k=0;k<prefiltered->h();k++) {
    for (int l=0;l<prefiltered->w();l++) {
      int ti = tileIds[(l/2+k/2*TILEMAP_SIZE)*4];
      ForIndex(c,3) {
        prefiltered->pixel(l,k)[c] = tex2dData[ti].back()[c];
      }
      prefiltered->pixel(l,k)[3] = 255;
    }
  }
  uint num_miplvls = 0;
  uint tmsz        = prefiltered->w();
  while (tmsz > 0) { // compute required number of levels
    tmsz = tmsz >> 1;
    num_miplvls ++;
  }
  // -> create a MIP-map pyramid for the prefiltered tilemap (simple box filter)
  vector<ImageRGBA*> pftm_pyr;
  pftm_pyr.resize(num_miplvls);
  pftm_pyr[0] = prefiltered;
  ForRange(l,1,pftm_pyr.size()-1) {
    pftm_pyr[l] = new ImageRGBA(max(1,prefiltered->w()>>l),max(1,prefiltered->h()>>l));
    ForImage(pftm_pyr[l],x,y) {
      v4f c00 = v4f(pftm_pyr[l-1]->pixel(x*2  ,y*2  ));
      v4f c01 = v4f(pftm_pyr[l-1]->pixel(x*2  ,y*2+1));
      v4f c11 = v4f(pftm_pyr[l-1]->pixel(x*2+1,y*2+1));
      v4f c10 = v4f(pftm_pyr[l-1]->pixel(x*2+1,y*2  ));
      v4f avg = (c00+c01+c10+c11)*0.25f;
      pftm_pyr[l]->pixel(x,y)    = v4b(avg);
      pftm_pyr[l]->pixel(x,y)[3] = 0;
    }
  }
  // -> data
  vector<D3D10_SUBRESOURCE_DATA> initDataPF;
  initDataPF.resize(num_miplvls);
  ForIndex(l,num_miplvls) {
    initDataPF[l].pSysMem              = pftm_pyr[l]->raw();
    initDataPF[l].SysMemPitch          = pftm_pyr[l]->w()*4;
    initDataPF[l].SysMemSlicePitch     = 0;
  }
  // -> dx10 description
  tex2dDesc.ArraySize			      = 1;
  tex2dDesc.Format			        = DXGI_FORMAT_R8G8B8A8_UNORM;
  tex2dDesc.Width				        = pftm_pyr[0]->w();
  tex2dDesc.Height			        = pftm_pyr[0]->h();
  tex2dDesc.MipLevels			      = num_miplvls;
  tex2dDesc.MiscFlags			      = 0;
  tex2dDesc.SampleDesc.Count    = 1;
  tex2dDesc.SampleDesc.Quality  = 0;
  tex2dDesc.CPUAccessFlags	    = 0;
  tex2dDesc.Usage				        = D3D10_USAGE_DEFAULT;
  tex2dDesc.BindFlags			      = D3D10_BIND_SHADER_RESOURCE;
  SAFE_RELEASE(g_PFTileMap);
  CHECKERR( pd3dDevice->CreateTexture2D( &tex2dDesc, &(initDataPF[0]), &g_PFTileMap ) );
  // -> dx10 view
  tex2dRVDesc.Format                    = DXGI_FORMAT_R8G8B8A8_UNORM;
  tex2dRVDesc.ViewDimension             = D3D10_SRV_DIMENSION_TEXTURE2D;
  tex2dRVDesc.Texture2D.MipLevels       = num_miplvls;
  tex2dRVDesc.Texture2D.MostDetailedMip = 0;
  SAFE_RELEASE(g_PFTileMapRV);
  CHECKERR( pd3dDevice->CreateShaderResourceView( g_PFTileMap, &tex2dRVDesc, &g_PFTileMapRV ) );

  // done.
  ForIndex(p,pyr.size()) {
    delete (pyr[p]);
  }
  ForIndex(p,pyr.size()) {
    delete (pftm_pyr[p]);
  }
}



//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
// All functions beyond this point are standard DXUT / DX10
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------



//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------

bool    CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );
void    CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void    CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void    CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void    CALLBACK OnMouse(bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos, void* pUserContext);
bool    CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void    CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext );
void    CALLBACK OnD3D10DestroyDevice( void* pUserContext );
void    CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );

void    InitApp();
void    RenderText();

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------

int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
  // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
  _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

  // Set DXUT callbacks
  InitApp();
  DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
  DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen

  DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
  DXUTSetCallbackMsgProc( MsgProc );
  DXUTSetCallbackKeyboard( OnKeyboard );
  DXUTSetCallbackFrameMove( OnFrameMove );
  DXUTSetCallbackMouse( OnMouse , true );

  DXUTSetCallbackD3D10DeviceAcceptable  ( IsD3D10DeviceAcceptable );
  DXUTSetCallbackD3D10SwapChainResized  ( OnD3D10ResizedSwapChain );
  DXUTSetCallbackD3D10FrameRender       ( OnD3D10FrameRender );
  DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10ReleasingSwapChain );
  DXUTSetCallbackD3D10DeviceDestroyed   ( OnD3D10DestroyDevice );
  DXUTSetCallbackD3D10DeviceCreated     ( OnD3D10CreateDevice );

  DXUTCreateWindow( L"ShaderX6 - Filtered Tilemaps demo" );
  DXUTCreateDevice( true, g_Width, g_Height );

  DXUTMainLoop(); // Enter into the DXUT render loop

  return DXUTGetExitCode();
}


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
  // Initialize dialogs
  g_D3DSettingsDlg.Init( &g_DialogResourceManager );
  g_HUD           .Init( &g_DialogResourceManager );
  g_SampleUI      .Init( &g_DialogResourceManager );

  g_HUD.SetCallback( OnGUIEvent ); int iY = 10; 
  g_HUD.AddButton ( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
  g_HUD.AddButton ( IDC_TOGGLEREF       , L"Toggle REF (F3)", 35, iY += 24, 125, 22, VK_F3 );
  g_HUD.AddButton ( IDC_RELOAD_SHADER   , L"Reload shader (F5)", 35, iY += 24, 125, 22, VK_F5 );
  g_HUD.AddListBox( IDC_RENDER_MODE     , 35, iY += 90 ,125, 70, 0, &g_lbRenderMode);
  g_HUD.AddListBox( IDC_TILE_SET        , 35, iY += 150,125,120, 0, &g_lbTileSet);
  
  g_SampleUI.SetCallback( OnGUIEvent );

  // init trackball
  trackballInit(g_Width,g_Height);
}

//--------------------------------------------------------------------------------------

bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
  if ( DXUT_D3D9_DEVICE == pDeviceSettings->ver ) {
    return false;
  }

  // For the first device created if its a REF device, optionally display a warning dialog box
  static bool s_bFirstTime = true;
  if( s_bFirstTime )
  {
    s_bFirstTime = false;
    if( (DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF) ||
      (DXUT_D3D10_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE) )
      DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
  }

  return true;
}


//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{

}

//--------------------------------------------------------------------------------------
// OnMouse
//--------------------------------------------------------------------------------------
void    CALLBACK OnMouse(bool bLeftButtonDown, 
                         bool bRightButtonDown, 
                         bool bMiddleButtonDown, 
                         bool bSideButton1Down, 
                         bool bSideButton2Down, 
                         int nMouseWheelDelta, 
                         int xPos, int yPos, 
                         void* pUserContext)
{

  static bool leftButtonWasDown   = false;
  static bool rightButtonWasDown  = false;
  static bool middleButtonWasDown = false;

  if (bLeftButtonDown && bRightButtonDown) {
    // emulate middle button
    bMiddleButtonDown = true;
    bLeftButtonDown   = false;
    bRightButtonDown  = false;
  }

  if (bLeftButtonDown && !leftButtonWasDown) {
    trackballButtonPressed(LEFT_BUTTON,xPos,yPos);
  } else if (!bLeftButtonDown && leftButtonWasDown) {
    trackballButtonReleased(LEFT_BUTTON);
  }
  leftButtonWasDown = bLeftButtonDown;

  if (bRightButtonDown && !rightButtonWasDown) {
    trackballButtonPressed(RIGHT_BUTTON,xPos,yPos);
  } else if (!bRightButtonDown && rightButtonWasDown) {
    trackballButtonReleased(RIGHT_BUTTON);
  }
  rightButtonWasDown = bRightButtonDown;

  if (bMiddleButtonDown && !middleButtonWasDown) {
    trackballButtonPressed(MIDDLE_BUTTON,xPos,yPos);
  } else if (!bMiddleButtonDown && middleButtonWasDown) {
    trackballButtonReleased(MIDDLE_BUTTON);
  }
  middleButtonWasDown = bMiddleButtonDown;

  // update trackball
  trackballUpdate(xPos,yPos);
}

//--------------------------------------------------------------------------------------
// Render the help and statistics text
//--------------------------------------------------------------------------------------
void RenderText()
{
  g_pTxtHelper->Begin();
  g_pTxtHelper->SetInsertionPos( 2, 0 );
  g_pTxtHelper->SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
  g_pTxtHelper->DrawTextLine( DXUTGetFrameStats( true ) );
  // g_pTxtHelper->DrawTextLine( DXUTGetDeviceStats() );    
  g_pTxtHelper->End();
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
  // Pass messages to dialog resource manager calls so GUI state is updated correctly
  *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
  if( *pbNoFurtherProcessing )
    return 0;

  // Pass messages to settings dialog if its active
  if( g_D3DSettingsDlg.IsActive() )
  {
    g_D3DSettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
    return 0;
  }

  // Give the dialogs a chance to handle the message first
  *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
  if( *pbNoFurtherProcessing )
    return 0;
  *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
  if( *pbNoFurtherProcessing )
    return 0;

  return 0;
}


//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
  if( bKeyDown )
  {
    switch( nChar )
    {
    case VK_F1:  break;
    }
  }
}

//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
  return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC *pBackBufferSurfaceDesc, void* pUserContext )
{

  CHECKERR( g_DialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
  CHECKERR( g_D3DSettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );
  CHECKERR( D3DX10CreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, 
    OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
    L"Arial", &g_pFont10 ) );
  CHECKERR( D3DX10CreateSprite( pd3dDevice, 512, &g_pSprite10 ) );
  g_pTxtHelper = new CDXUTTextHelper( NULL, NULL, g_pFont10, g_pSprite10, 15 );

  //////////////////////////////////////////////////////////////////
  
  D3D10_BLEND_DESC BlendState;
  ZeroMemory(&BlendState, sizeof(D3D10_BLEND_DESC));
  BlendState.BlendEnable[0]           = FALSE;
  BlendState.RenderTargetWriteMask[0] = D3D10_COLOR_WRITE_ENABLE_ALL;
  pd3dDevice->CreateBlendState(&BlendState, &g_StateNoBlend);

  //////////////////////////////////////////////////////////////////

  loadShader(pd3dDevice);

  //////////////////////////////////////////////////////////////////

  // Create a quad

  // VL
  const D3D10_INPUT_ELEMENT_DESC quad_layout[] =
  {
    { "POSITION",  0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
    { "TEXCOORD",  0, DXGI_FORMAT_R32G32_FLOAT,    0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
  };

  D3D10_PASS_DESC passDesc;
  CHECKERR( g_RenderTMTechnique->GetPassByIndex( 0 )->GetDesc( &passDesc ) );
  CHECKERR( pd3dDevice->CreateInputLayout( quad_layout, 2, passDesc.pIAInputSignature, passDesc.IAInputSignatureSize, &g_QuadVL ) );

  // VB
  quad_vertex quad[] =
  {
    D3DXVECTOR3(-1.0f,-1.0f, 0.0f ), D3DXVECTOR2( 0.0f, 1.0f ),
    D3DXVECTOR3(-1.0f, 1.0f, 0.0f ), D3DXVECTOR2( 0.0f, 0.0f ),
    D3DXVECTOR3( 1.0f, 1.0f, 0.0f ), D3DXVECTOR2( 1.0f, 0.0f ),
    D3DXVECTOR3( 1.0f,-1.0f, 0.0f ), D3DXVECTOR2( 1.0f, 1.0f ),
  };

  D3D10_BUFFER_DESC bufferDesc;
  bufferDesc.Usage 			= D3D10_USAGE_DEFAULT;
  bufferDesc.ByteWidth 		= sizeof( quad_vertex ) * 4;
  bufferDesc.BindFlags 		= D3D10_BIND_VERTEX_BUFFER;
  bufferDesc.CPUAccessFlags 	= 0;
  bufferDesc.MiscFlags 		= 0;

  D3D10_SUBRESOURCE_DATA initData;
  initData.pSysMem          = quad;
  initData.SysMemPitch      = 0;
  initData.SysMemSlicePitch = 0;
  CHECKERR( pd3dDevice->CreateBuffer( &bufferDesc, &initData, &g_QuadVB ) );

  // IB
  unsigned int indices[]     = { 0, 1, 2, 0, 2, 3 };

  bufferDesc.Usage           = D3D10_USAGE_DEFAULT;
  bufferDesc.ByteWidth       = sizeof( unsigned int ) * 6;
  bufferDesc.BindFlags 		   = D3D10_BIND_INDEX_BUFFER;
  bufferDesc.CPUAccessFlags  = 0;
  bufferDesc.MiscFlags 		   = 0;

  initData.pSysMem            = indices;
  initData.SysMemPitch        = 0;
  initData.SysMemSlicePitch   = 0;
  CHECKERR( pd3dDevice->CreateBuffer( &bufferDesc, &initData, &g_QuadIB ) );

  //////////////////////////////////////////////////////////////////

  loadTiles(pd3dDevice,"4x4.pipes.png");

  //////////////////////////////////////////////////////////////////

  setShaderParameters(pd3dDevice);

  //////////////////////////////////////////////////////////////////

  return S_OK;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
  HRESULT hr;

  V_RETURN( g_DialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
  V_RETURN( g_D3DSettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

  g_Width  = pBackBufferSurfaceDesc->Width;
  g_Height = pBackBufferSurfaceDesc->Height;

  trackballInit(g_Width,g_Height);

  g_HUD.SetLocation( pBackBufferSurfaceDesc->Width-170, 0 );
  g_HUD.SetSize( 170, 170 );
  g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width-170, pBackBufferSurfaceDesc->Height-300 );
  g_SampleUI.SetSize( 170, 300 );

  return S_OK;
}


//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
  // If the settings dialog is being shown, then render it instead of rendering the app's scene
  if( g_D3DSettingsDlg.IsActive() )
  {
    g_D3DSettingsDlg.OnRender( fElapsedTime );
    return;
  }

  // Clear the render target and depth stencil
  float ClearColor[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
  ID3D10RenderTargetView* pRTV = DXUTGetD3D10RenderTargetView();
  ID3D10DepthStencilView* pDSV = DXUTGetD3D10DepthStencilView();
  pd3dDevice->ClearRenderTargetView( pRTV, ClearColor );
  pd3dDevice->ClearDepthStencilView( pDSV, D3D10_CLEAR_DEPTH, 1.0, 0 );

  // No blending
  pd3dDevice->OMSetBlendState(g_StateNoBlend,0,0xffffffff);

  D3DXMATRIX prj;
  D3DXMatrixPerspectiveFovRH(&prj,3.14f/4.0f,g_Width/float(g_Height),0.01f,100.0f);
  D3DXMatrixTranspose(&prj,&prj);
  g_fxProj->SetMatrix(&prj[0]);

  // Set trackball
  m4x4f mat    = trackballMatrix();
  mat          = mat.transpose();
  g_fxTrackball->SetMatrix(&mat[0]);

  // Render the quad
  // Set the input layout
  pd3dDevice->IASetInputLayout( g_QuadVL );
  // Set VB
  UINT stride = sizeof( quad_vertex );
  UINT offset = 0;
  pd3dDevice->IASetVertexBuffers( 
    0,                // the first input slot for binding
    1,                // the number of buffers in the array
    &g_QuadVB,        // the array of vertex buffers
    &stride,          // array of stride values, one for each buffer
    &offset );        // array of offset values, one for each buffer
  // Set IB
  pd3dDevice->IASetIndexBuffer( g_QuadIB, DXGI_FORMAT_R32_UINT, 0 );
  // Set topology
  pd3dDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
  // Draw
  g_SelectedTechnique->GetPassByIndex(0)->Apply(0);
  pd3dDevice->DrawIndexed( 6, 0, 0 );

  // Render the HUD
  pd3dDevice->ClearDepthStencilView( pDSV, D3D10_CLEAR_DEPTH, 1.0, 0 );
  DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"HUD / Stats" );
  g_HUD     .OnRender( fElapsedTime ); 
  //  g_SampleUI.OnRender( fElapsedTime );
  RenderText();
  DXUT_EndPerfEvent();

}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext )
{
  g_DialogResourceManager.OnD3D10ReleasingSwapChain();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------

void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
  g_DialogResourceManager.OnD3D10DestroyDevice();
  g_D3DSettingsDlg.OnD3D10DestroyDevice();
  CDXUTDirectionWidget::StaticOnD3D10DestroyDevice();
  DXUTGetGlobalResourceCache().OnDestroyDevice();

  SAFE_DELETE ( g_pTxtHelper );
  SAFE_RELEASE( g_pFont10 );
  SAFE_RELEASE( g_pSprite10 );

  SAFE_RELEASE( g_StateNoBlend );

  SAFE_RELEASE( g_QuadIB );
  SAFE_RELEASE( g_QuadVB );
  SAFE_RELEASE( g_QuadVL );

  SAFE_RELEASE( g_Array2D );
  SAFE_RELEASE( g_Array2DRV );
  SAFE_RELEASE( g_Tex2D );
  SAFE_RELEASE( g_Tex2DRV );
  SAFE_RELEASE( g_TileMap );
  SAFE_RELEASE( g_TileMapRV );
  SAFE_RELEASE( g_PFTileMap );
  SAFE_RELEASE( g_PFTileMapRV );

  SAFE_RELEASE( g_pEffect10 );

}

//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------

void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{   
  switch( nControlID )
  {

  case IDC_TOGGLEFULLSCREEN: 
    DXUTToggleFullScreen(); 
    break;

  case IDC_TOGGLEREF:        
    DXUTToggleREF(); 
    break;

  case IDC_RELOAD_SHADER: 
    loadShader         (DXUTGetD3D10Device());
    setShaderParameters(DXUTGetD3D10Device());
    break;

  case IDC_RENDER_MODE:
    g_SelectedTechnique = (ID3D10EffectTechnique*)(g_lbRenderMode->GetSelectedItem()->pData);
    break;

  case IDC_TILE_SET:
    loadTiles          (DXUTGetD3D10Device(),(const char *)g_lbTileSet->GetSelectedItem()->pData);
    setShaderParameters(DXUTGetD3D10Device());
    break;

  }
}

//--------------------------------------------------------------------------------------
