/*

Demo code for 'Filtered Tilemaps' - ShaderX6

--

This code comes as it is, with no warranty. Use at your own risk.
Permission to reuse the code in this file is granted.

--

Sylvain Lefebvre - 2007-03-29

*/

uniform float TileWidth;
uniform float TileWidth_inv;

static const float TileMapSize   = 128.0;

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

struct PS_OUTPUT
{
    float4 Color : SV_Target;  // Pixel color
};

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

struct VS_QUAD_OUTPUT
{
    float4 Pos       : SV_POSITION; // vertex position 
    float2 UV        : TEXCOORD0;   // texture coords 
    float2 TileMapUV : TEXCOORD1;   // tilemap coords 
};

float4x4 Proj;
float4x4 Trackball;

VS_QUAD_OUTPUT RenderQuadVS( float4 vPos : POSITION,
                             float2 vUV  : TEXCOORD)
{
    VS_QUAD_OUTPUT output;

    output.Pos = mul(Proj, mul(Trackball , float4(vPos.xy,0,1))); // on screen quad
    
    output.UV        = vUV;
    output.TileMapUV = vUV * TileMapSize;

    return output;    
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

SamplerState NearestSampler
{
    Filter   = MIN_MAG_MIP_POINT;
    AddressU = Wrap;
    AddressV = Wrap;
};

SamplerState LinearSampler
{
    Filter   = MIN_MAG_MIP_LINEAR;
    AddressU = Border;
    AddressV = Border;
};

SamplerState AnisoSampler
{
    Filter    = ANISOTROPIC;
    AddressU  = Border;
    AddressV  = Border;
};

SamplerState WrapLinearSampler
{
    Filter   = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};

SamplerState WrapAnisoSampler
{
    Filter    = ANISOTROPIC;
    AddressU  = Wrap;
    AddressV  = Wrap;
};

#define USE_SAMPLER AnisoSampler
// #define USE_SAMPLER LinearSampler

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

Texture2D     <uint4 > TileMap;     // Tilemap
Texture2DArray<float4> Array2d;     // Array of tiles
Texture2D     <float4> Packed;      // Packed tiles texture
Texture2D     <float4> PreFiltered; // Pre filtered tilemap for distant views

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Naive lookup

PS_OUTPUT RenderTileMapPS( VS_QUAD_OUTPUT In ) 
{ 
    PS_OUTPUT output;

	// get tile id
	float  id    = TileMap.Load(int3(In.TileMapUV,0));
	// compute uv within the tile
	float2 uv    = frac(In.TileMapUV);

	float2 duvdx = ddx(In.TileMapUV);
	float2 duvdy = ddy(In.TileMapUV);
	output.Color = Array2d.SampleGrad(USE_SAMPLER,float3(uv,id),duvdx,duvdy);

    return output;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Lookup using packed tiles texture and continuous packing

static const uint P = 4;

PS_OUTPUT RenderPackedTileMapPS( VS_QUAD_OUTPUT In ) 
{ 
    PS_OUTPUT output;

    // get tile id
    uint   id    = TileMap.Load(int3(In.TileMapUV,0));
	
    // compute uv
    float2 uv    = (frac(In.TileMapUV) + float2(id%P,id/P)) / float(P);

    // compute correct derivatives
    float2 duvdx = ddx(In.TileMapUV / float(P));
    float2 duvdy = ddy(In.TileMapUV / float(P));
    output.Color = Packed.SampleGrad(WrapAnisoSampler,uv,duvdx,duvdy);

    return output;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

float2 tileCoord(float2 uvtile, out uint2 ij)
{
    // tile coordinate in the tiling
    ij      = uint2(uvtile);
    // local uv inside tile
    uvtile  = uvtile - ij;
    return (uvtile);
}

// ----------

// Lookup with new interpolation scheme

PS_OUTPUT RenderInterpolatedTileMapPS( VS_QUAD_OUTPUT In ) 
{ 
    PS_OUTPUT output;

	float2 duvdx = ddx(In.TileMapUV);
	float2 duvdy = ddy(In.TileMapUV);

#if 0
	
	// code path for standard tilemap

	uint2  ij    = 0;
	float2 tc00  = tileCoord(In.TileMapUV,ij);

	float4 tcs;
	uint4  ijs;

	tcs          = float4(tc00,tc00 > 0.5 ? tc00 - 1.0 : tc00 + 1.0);
	ijs          = uint4 (ij  ,tc00 > 0.5 ? ij   + 1   : ij   - 1  );

	float2  tc10 = tcs.zy;
	float2  tc11 = tcs.zw;
	float2  tc01 = tcs.xw;

	uint    id00 = TileMap.Load(int3(ij,0));
	uint    id10 = TileMap.Load(int3(ijs.zy,0));
	uint    id11 = TileMap.Load(int3(ijs.zw,0));
	uint    id01 = TileMap.Load(int3(ijs.xw,0));

	output.Color = 
	    Array2d.SampleGrad(USE_SAMPLER,float3(tc00,id00),duvdx,duvdy)
	  + Array2d.SampleGrad(USE_SAMPLER,float3(tc10,id10),duvdx,duvdy)
	  + Array2d.SampleGrad(USE_SAMPLER,float3(tc11,id11),duvdx,duvdy)
	  + Array2d.SampleGrad(USE_SAMPLER,float3(tc01,id01),duvdx,duvdy);

#else

	// code path for tilemap storing tile id of neighbors (single lookup into the tilemap)

	float4 pf_clr    = PreFiltered.Sample(USE_SAMPLER,In.UV);

	uint2  ij        = 0;
	float2 tc        = tileCoord(In.TileMapUV - 0.5,ij);

    float2  tc00     = tc   + 0.5;
    float2  tc11     = tc00 - 1.0;
    const float2 z10 = float2(1.0,0.0);
    float2  tc10     = tc00 - z10;
    float2  tc01     = tc00 - z10.yx;
    
	uint4   ids      = TileMap.Load(int3(ij,0));

	float4 clr       = 
	    Array2d.SampleGrad(USE_SAMPLER,float3(tc00,ids.x),duvdx,duvdy)
	  + Array2d.SampleGrad(USE_SAMPLER,float3(tc10,ids.y),duvdx,duvdy)
	  + Array2d.SampleGrad(USE_SAMPLER,float3(tc11,ids.w),duvdx,duvdy)
	  + Array2d.SampleGrad(USE_SAMPLER,float3(tc01,ids.z),duvdx,duvdy);
	  
	output.Color     = lerp(pf_clr,clr,pf_clr.w*clr.w);

#endif
	
    return output;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Lookup with new interpolation scheme. Uses dynamic branching to avoid
// interpolation between tiles away from boundaries.

PS_OUTPUT RenderInterpolatedTileMapConditionalPS( VS_QUAD_OUTPUT In ) 
{ 
    PS_OUTPUT output;

	float4 pf_clr    = PreFiltered.Sample(USE_SAMPLER,In.UV);
	float4 clr;

	/*if (pf_clr.w > 0)*/ {

		uint2  ij      = 0;
		float2 tc00    = tileCoord(In.TileMapUV,ij);

		float2 duvdx   = ddx(In.TileMapUV);
		float2 duvdy   = ddy(In.TileMapUV);
		
		float2 d       = abs(tc00-0.5);
		if (max(d.x,d.y) < 0.5 - max(TileWidth_inv,max(length(duvdx),length(duvdy)))) {
		  
		  uint    id00 = TileMap.Load(int3(ij,0));
		  clr          = Array2d.SampleGrad(USE_SAMPLER,float3(tc00,id00),duvdx,duvdy);
		  
		} else {

#if 0
		  // code path for standard tilemap

		  float4 tcs;
		  uint4  ijs;
			
		  tcs           = float4(tc00,tc00 > 0.5 ? tc00 - 1.0 : tc00 + 1.0);
		  ijs           = uint4 (ij  ,tc00 > 0.5 ? ij   + 1   : ij   - 1  );

		  float2  tc10  = tcs.zy;
		  float2  tc11  = tcs.zw;
		  float2  tc01  = tcs.xw;

		  uint    id00  = TileMap.Load(int3(ij,0));
		  uint    id10  = TileMap.Load(int3(ijs.zy,0));
		  uint    id11  = TileMap.Load(int3(ijs.zw,0));
		  uint    id01  = TileMap.Load(int3(ijs.xw,0));

#else 

		  // code path for tilemap storing tile id of neighbors (single lookup into the tilemap)

		  tc00          = tileCoord(In.TileMapUV - 0.5,ij);
		  tc00          = tc00 + 0.5;
		  float2  tc10  = tc00 - float2(1.0,0.0);
		  float2  tc11  = tc00 - float2(1.0,1.0);
		  float2  tc01  = tc00 - float2(0.0,1.0);
	    
		  uint4   ids   = TileMap.Load(int3(ij,0));
		  uint    id00  = ids.x;
		  uint    id10  = ids.y;
		  uint    id01  = ids.z;
		  uint    id11  = ids.w;

#endif	
	
		  clr  = 
		      Array2d.SampleGrad(USE_SAMPLER,float3(tc00,id00),duvdx,duvdy)
		    + Array2d.SampleGrad(USE_SAMPLER,float3(tc10,id10),duvdx,duvdy)
		    + Array2d.SampleGrad(USE_SAMPLER,float3(tc11,id11),duvdx,duvdy)
		    + Array2d.SampleGrad(USE_SAMPLER,float3(tc01,id01),duvdx,duvdy);
		}
    }
    
   	output.Color     = lerp(pf_clr,clr,pf_clr.w*clr.w);

    return output;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

technique10 RenderTileMap
{
    pass P0
    {
        SetVertexShader  ( CompileShader( vs_4_0, RenderQuadVS( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, RenderTileMapPS( ) ) );
    }
}

//--------------------------------------------------------------------------------------

technique10 RenderInterpolatedTileMap
{
    pass P0
    {
        SetVertexShader  ( CompileShader( vs_4_0, RenderQuadVS( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, RenderInterpolatedTileMapPS( ) ) );
    }
}

//--------------------------------------------------------------------------------------

technique10 RenderInterpolatedTileMapConditional
{
    pass P0
    {
        SetVertexShader  ( CompileShader( vs_4_0, RenderQuadVS( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, RenderInterpolatedTileMapConditionalPS( ) ) );
    }
}

//--------------------------------------------------------------------------------------

technique10 RenderPackedTileMap
{
    pass P0
    {
        SetVertexShader  ( CompileShader( vs_4_0, RenderQuadVS( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, RenderPackedTileMapPS( ) ) );
    }
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
